#include <stdlib.h>
#include <string.h>

#include <nllibc.h>

struct area_header {
  struct area_header *next;
  struct area_header *prev;
  int size;
  int magic;
#define AREA_HEADER_MAGIC_NUMBER 0x12345678
};

static struct area_header *alloc_start = NULL;
static struct area_header *alloc_end = NULL;

static char space[1024 * 1024];

static int init()
{
  if (alloc_start == NULL) {
    alloc_start = (struct area_header *)space;
    alloc_end = (struct area_header *)(space + sizeof(space)) - 1;
    alloc_start->prev = NULL;
    alloc_start->next = alloc_end;
    alloc_end->prev = alloc_start;
    alloc_end->next = NULL;
    alloc_start->size = 0;
    alloc_end->size = 0;
  }
  return 0;
}

void *malloc(size_t size)
{
  struct area_header *area, *following;
  int aligned_size;

  init();

  aligned_size = ((size + 3) / 4) * 4;

  for (area = alloc_start; area && area->next; area = area->next) {
    following = (struct area_header *)((char *)(area + 1) + area->size);
    if ((char *)area->next >= (char *)(following + 1) + aligned_size) {
      following->size = aligned_size;
      following->magic = AREA_HEADER_MAGIC_NUMBER;
      following->prev = area;
      following->next = area->next;
      area->next->prev = following;
      area->next = following;
      return following + 1;
    }
  }

  return NULL;
}

void *calloc(size_t number, size_t size)
{
  int s = size * number;
  void *p;

  p = malloc(s);
  if (p != NULL) {
    memset(p, 0, s);
  }

  return p;
}

void *realloc(void *ptr, size_t size)
{
  struct area_header *area;
  int aligned_size;
  void *p;

  if (ptr != NULL) {
    aligned_size = ((size + 3) / 4) * 4;

    area = ((struct area_header *)ptr) - 1;
    if (aligned_size <= area->size)
      return ptr;

    if ((char *)area->next >= (char *)(area + 1) + aligned_size) {
      area->size = aligned_size;
      return ptr;
    }
  }

  p = malloc(size);
  if (p == NULL)
    return NULL;

  if (ptr) {
    memcpy(p, ptr, size);
    free(ptr);
  }

  return p;
}

void free(void *ptr)
{
  struct area_header *area;

  if (ptr == NULL)
    return;

  area = ((struct area_header *)ptr) - 1;

  if (area->magic != AREA_HEADER_MAGIC_NUMBER) {
    ;
  }

  area->prev->next = area->next;
  area->next->prev = area->prev;
  area->magic = 0;

  return;
}
